@@ -43,6 +43,7 @@
 #include <stdatomic.h>
 #include <limits.h>        /* PATH_MAX */
 #include <locale.h>
+#include <pthread.h>
 #ifndef HAVE_GETTID
 #include <sys/syscall.h>
 #endif
@@ -63,31 +64,38 @@
 #include "gcc-attributes.h"
 #include "avl.h"
 #include "paths.h"
-
+#include "string-util.h"
+#include "filter.h"
 
 // Global program variables
-unsigned int debug_mode = 0, permissive = 0;
-const char* mounts = "/proc/mounts";
+unsigned int debug_mode = 0;
+const char* mounts = MOUNTS_FILE;
 
 // Signal handler notifications
 atomic_bool stop = false, hup = false, run_stats = false;
 
-// Local variables
-static conf_t config;
+// Global configuration state
+conf_t config;
 // This holds info about all file systems to watch
 struct fs_avl {
 	avl_tree_t index;
 };
 // This is the data about a specific file system to watch
 typedef struct fs_data {
-        avl_t avl;        // This has to be first
-        const char *fs_name;
+	avl_t avl;        // This has to be first
+	const char *fs_name;
 } fs_data_t;
 static struct fs_avl filesystems;
+static struct fs_avl ignored_mounts;
 
 // List of mounts being watched
 static mlist *m = NULL;
 
+// Reconfiguration
+static atomic_bool reconfig_running = false;
+// Mount handling
+static atomic_bool mounts_running = false;
+
 static void usage(void) NORETURN;
 
 #ifndef HAVE_GETTID
@@ -139,35 +147,52 @@ static int cmp_fs(void *a, void *b)
 }
 
 
+/*
+ * free_filesystem - release storage for a single AVL entry.
+ * @s: filesystem data being destroyed.
+ * Returns nothing.
+ */
 static void free_filesystem(fs_data_t *s)
 {
 	free((void *)s->fs_name);
 	free((void *)s);
 }
 
-
-static void destroy_filesystem(void)
+/*
+ * destroy_filesystem - remove the first node from an AVL tree.
+ * @list: tree holding filesystem style strings.
+ * Returns nothing.
+ */
+static void destroy_filesystem(struct fs_avl *list)
 {
-	avl_t *cur = filesystems.index.root;
+	avl_t *cur = list->index.root;
 
-	fs_data_t *tmp =(fs_data_t *)avl_remove(&filesystems.index, cur);
+	fs_data_t *tmp =(fs_data_t *)avl_remove(&list->index, cur);
 	if ((avl_t *)tmp != cur)
-		msg(LOG_DEBUG, "filesystem: removal of invalid node");
+		msg(LOG_DEBUG, "string list: removal of invalid node");
 	free_filesystem(tmp);
 }
 
-
-static void destroy_fs_list(void)
+/*
+ * destroy_fs_list - free every node from an AVL tree.
+ * @list: tree that should be cleared.
+ * Returns nothing.
+ */
+static void destroy_fs_list(struct fs_avl *list)
 {
-	while (filesystems.index.root)
-		destroy_filesystem();
+	while (list->index.root)
+		destroy_filesystem(list);
 }
 
-// Add a filesystem to the AVL tree.
-// Returns: 0 on failure, 1 if item is in the AVL tree.
-static int add_filesystem(fs_data_t *f)
+/*
+ * add_filesystem - insert a string into an AVL tree.
+ * @list: tree receiving the entry.
+ * @f: filesystem data to insert.
+ * Returns: 0 on failure, 1 if the item is in the tree.
+ */
+static int add_filesystem(struct fs_avl *list, fs_data_t *f)
 {
-	fs_data_t *tmp=(fs_data_t *)avl_insert(&filesystems.index,(avl_t *)(f));
+	fs_data_t *tmp=(fs_data_t *)avl_insert(&list->index,(avl_t *)(f));
 	if (tmp) {
 		if (tmp != f) {
 			// already in the tree, delete the current item
@@ -179,14 +204,18 @@ static int add_filesystem(fs_data_t *f)
 	return 0;
 }
 
-// Create fsdata_t struct and add it to the filesystem AVL tree
-// Returns 1 on success and 0 on failure
-static int new_filesystem(const char *fs)
+/*
+ * new_filesystem - allocate and add a string into an AVL tree.
+ * @list: tree receiving the new entry.
+ * @fs: string representing the filesystem or mount point.
+ * Returns 1 on success and 0 on failure.
+ */
+static int new_filesystem(struct fs_avl *list, const char *fs)
 {
 	fs_data_t *tmp = malloc(sizeof(fs_data_t));
 	if (tmp) {
 		tmp->fs_name = fs ? strdup(fs) : strdup("");
-		if (add_filesystem(tmp) == 0) {
+		if (add_filesystem(list, tmp) == 0) {
 			free((void *)tmp->fs_name);
 			free(tmp);
 			return 0;
@@ -196,16 +225,80 @@ static int new_filesystem(const char *fs)
 	return 0;
 }
 
-
-static fs_data_t *find_filesystem(const char *f)
+/*
+ * find_filesystem - search the supplied AVL tree for a string.
+ * @list: tree that should be searched.
+ * @f: string to locate.
+ * Returns a pointer to the stored entry or NULL if not found.
+ */
+static fs_data_t *find_filesystem(struct fs_avl *list, const char *f)
 {
 	fs_data_t tmp;
 
 	tmp.fs_name = f;
-	return (fs_data_t *)avl_search(&filesystems.index, (avl_t *) &tmp);
+	return (fs_data_t *)avl_search(&list->index, (avl_t *) &tmp);
 }
 
+/*
+ * add_ignore_mount_entry - callback that validates and stores ignored mounts.
+ * @mount: trimmed ignore_mounts entry.
+ * @unused: unused context pointer.
+ * Returns 0 to continue iterating entries.
+ */
+static int add_ignore_mount_entry(const char *mount,
+	void *unused __attribute__ ((unused)))
+{
+	const char *warning;
+	int rc;
 
+	rc = check_ignore_mount_warning(mounts, mount, &warning);
+	if (warning)
+		msg(LOG_ERR, warning, mount, mounts);
+
+	if (rc == 1) {
+		if (!new_filesystem(&ignored_mounts, mount))
+			msg(LOG_ERR, "Cannot store ignore_mounts entry %s", mount);
+	}
+
+	return 0;
+}
+/*
+ * init_ignore_mounts - create the ignored mount AVL tree.
+ * @ignore_list: comma separated list of mount points to ignore.
+ * Returns nothing.
+ */
+static void init_ignore_mounts(const char *ignore_list)
+{
+	avl_init(&ignored_mounts.index, cmp_fs);
+
+	if (ignore_list == NULL)
+		return;
+
+	if (iterate_ignore_mounts(ignore_list, add_ignore_mount_entry, NULL)) {
+		msg(LOG_ERR, "Cannot duplicate ignore_mounts list");
+		return;
+	}
+
+	if (ignored_mounts.index.root == NULL) {
+		free((void *)config.ignore_mounts);
+		config.ignore_mounts = NULL;
+	}
+}
+/*
+ * mount_is_ignored - check if a mount point is in the ignore list.
+ * @point: mount point path.
+ * Returns 1 when the mount point should be skipped and 0 otherwise.
+ */
+static int mount_is_ignored(const char *point)
+{
+	return find_filesystem(&ignored_mounts, point) ? 1 : 0;
+}
+
+/*
+ * init_fs_list - create the filesystem type AVL tree.
+ * @watch_fs: comma separated list of filesystem types to watch.
+ * Returns nothing. Exits on failure when the list is missing.
+ */
 static void init_fs_list(const char *watch_fs)
 {
 	if (watch_fs == NULL) {
@@ -216,18 +309,24 @@ static void init_fs_list(const char *watch_fs)
 
 	// Now parse up list and push into avl
 	char *ptr, *saved, *tmp = strdup(watch_fs);
+
+	if (tmp == NULL) {
+		msg(LOG_ERR, "Cannot duplicate watch_fs list");
+		return;
+	}
+
 	ptr = strtok_r(tmp, ",", &saved);
 	while (ptr) {
-		new_filesystem(ptr);
+		new_filesystem(&filesystems, ptr);
 		ptr = strtok_r(NULL, ",", &saved);
 	}
 	free(tmp);
 }
 
-
 static void term_handler(int sig __attribute__((unused)))
 {
 	stop = true;
+	nudge_queue();
 }
 
 
@@ -258,19 +357,155 @@ static void hup_handler(int sig __attribute__((unused)))
 static void usr1_handler(int sig __attribute__((unused)))
 {
 	run_stats = true;
+	nudge_queue();
 }
 
 /*
- * This function handles the reconfiguration of the daemon
- * after receiving a SIGHUP signal.
+ * reload_configuration - refresh runtime configuration settings.
+ * @void: no arguments are required.
+ * Returns 0 when the configuration was reloaded, non-zero otherwise.
  */
+static int reload_configuration(void)
+{
+	conf_t new_config;
+
+	if (load_daemon_config(&new_config)) {
+		free_daemon_config(&new_config);
+		msg(LOG_ERR, "Failed reloading daemon configuration");
+		return 1;
+	}
+
+	config.permissive = new_config.permissive;
+
+	if (setpriority(PRIO_PROCESS, 0, -(int)new_config.nice_val) == -1)
+		msg(LOG_WARNING, "Couldn't adjust priority (%s)",
+		strerror(errno));
+	config.nice_val = new_config.nice_val;
+
+	config.do_stat_report = new_config.do_stat_report;
+	config.detailed_report = new_config.detailed_report;
+
+	if (new_config.integrity != config.integrity) {
+		set_integrity_mode(new_config.integrity);
+		config.integrity = new_config.integrity;
+	}
+
+	if (new_config.syslog_format &&
+		(!config.syslog_format ||
+		 strcmp(new_config.syslog_format, config.syslog_format) != 0)) {
+		char *new_syslog = strdup(new_config.syslog_format);
+		if (new_syslog) {
+			char *old_syslog;
+
+			lock_rule();
+			old_syslog = (char *)config.syslog_format;
+			config.syslog_format = new_syslog;
+			unlock_rule();
+
+			free(old_syslog);
+		} else
+			msg(LOG_ERR,
+				"Failed replacing syslog_format string");
+		}
+
+	config.rpm_sha256_only = new_config.rpm_sha256_only;
+
+	if (new_config.trust && (!config.trust ||
+				strcmp(new_config.trust, config.trust) != 0)) {
+		char *new_trust = strdup(new_config.trust);
+		if (new_trust) {
+			char *old_trust = (char *)config.trust;
+			config.trust = new_trust;
+			free(old_trust);
+		} else
+			msg(LOG_ERR,
+			    "Failed replacing trust backend list");
+	}
+
+	/*
+	 * Remaining daemon_config fields require restart-time changes:
+	 * q_size, subj_cache_size, and obj_cache_size are consumed when the
+	 * event queue and caches are created. uid/gid, allow_filesystem_mark,
+	 * watch_fs, and ignore_mounts are applied while fanotify marks are
+	 * installed. db_max_size fixes the LMDB map when the database opens,
+	 * and report_interval is bound to the decision thread's timer. None
+	 * of these components support resizing in-place yet, so their
+	 * configuration stays static.
+	 */
+
+	free_daemon_config(&new_config);
+
+	return 0;
+}
+
 static void reconfigure(void)
 {
+	if (reload_configuration())
+		msg(LOG_WARNING,
+			"Continuing with previous configuration settings");
+
+	filter_destroy();
+	if (filter_init())
+		msg(LOG_ERR, "Failed initializing filter configuration");
+	else if (filter_load_file(NULL))
+		msg(LOG_ERR, "Failed reloading filter configuration");
+
 	set_reload_rules();
-
 	set_reload_trust_database();
+}
 
-	// TODO: Update configuration
+/*
+ * reconfigure_thread_main - run configuration reload outside event loop.
+ * @arg: unused pointer.
+ * Returns NULL.
+ */
+static void *reconfigure_thread_main(void *arg  __attribute__ ((unused)))
+{
+	sigset_t sigs;
+
+	/* This is a worker thread. Don't handle external signals. */
+	sigemptyset(&sigs);
+	sigaddset(&sigs, SIGTERM);
+	sigaddset(&sigs, SIGHUP);
+	sigaddset(&sigs, SIGUSR1);
+	sigaddset(&sigs, SIGINT);
+	sigaddset(&sigs, SIGQUIT);
+	pthread_sigmask(SIG_SETMASK, &sigs, NULL);
+
+	reconfigure();
+	atomic_store(&reconfig_running, false);
+	return NULL;
+}
+
+/*
+ * maybe_start_reconfigure_thread - start a reconfigure thread when requested
+ * and prevent a second one running until first finishes.
+ * @void: no arguments are required.
+ * Returns nothing.
+ */
+static void maybe_start_reconfigure_thread(void)
+{
+	int rc;
+	bool expected = false;
+
+	// Make sure one is not runnning since it is detached
+	if (!atomic_compare_exchange_strong(&reconfig_running, &expected, true))
+		return;
+
+	// OK to start up the thread
+	pthread_t tid;
+	pthread_attr_t attr;
+	pthread_attr_init(&attr);
+	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
+
+	rc = pthread_create(&tid, &attr,
+			reconfigure_thread_main, NULL);
+	if (rc) {
+		msg(LOG_ERR, "Failed starting reconfigure thread (%s)",
+			strerror(rc));
+		atomic_store(&reconfig_running, false);
+	}
+	pthread_attr_destroy(&attr);
 }
 
 // This is a workaround for https://bugzilla.redhat.com/show_bug.cgi?id=643031
@@ -331,10 +566,13 @@ static int become_daemon(void)
 			}
 			if (dup2(fd, 1) < 0) {
 				close(fd);
+				close(0);
 				return -1;
 			}
 			if (dup2(fd, 2) < 0) {
 				close(fd);
+				close(0);
+				close(1);
 				return -1;
 			}
 			close(fd);
@@ -356,19 +594,20 @@ static int become_daemon(void)
 // Returns 1 if we care about the entry and 0 if we do not
 static int check_mount_entry(const char *point, const char *type)
 {
-	// Some we know we don't want
-	if (strcmp(point, "/run") == 0)
-		return 0;
-	if (strncmp(point, "/sys", 4) == 0)
+	// Skip entries explicitly ignored by configuration
+	if (mount_is_ignored(point))
 		return 0;
 
-	if (find_filesystem(type))
+	// Some we know we don't want
+	if (strncmp(point, "/sys/", 5) == 0)
+		return 0;
+
+	if (find_filesystem(&filesystems, type))
 		return 1;
 	else
 		return 0;
 }
 
-
 static void handle_mounts(int fd)
 {
 	char buf[PATH_MAX * 2], device[1025], point[4097];
@@ -401,7 +640,7 @@ static void handle_mounts(int fd)
 				// Can we find it in the old list?
 				if (mlist_find(m, point)) {
 					// Mark no change
-					m->cur->status = NO_CHANGE;
+					m->cur->status = MNT_NO_CHANGE;
 				} else
 					mlist_append(m, point);
 			}
@@ -414,12 +653,84 @@ static void handle_mounts(int fd)
 	fanotify_update(m);
 }
 
+/*
+ * handle_mounts_thread_main - run mount processing outside event loop.
+ * @arg: pointer to a heap-allocated file descriptor integer.
+ * Returns NULL.
+ */
+static void *handle_mounts_thread_main(void *arg)
+{
+	int fd;
+	sigset_t sigs;
+
+	/* This is a worker thread. Don't handle external signals. */
+	sigemptyset(&sigs);
+	sigaddset(&sigs, SIGTERM);
+	sigaddset(&sigs, SIGHUP);
+	sigaddset(&sigs, SIGUSR1);
+	sigaddset(&sigs, SIGINT);
+	sigaddset(&sigs, SIGQUIT);
+	pthread_sigmask(SIG_SETMASK, &sigs, NULL);
+
+	if (arg == NULL) {
+		atomic_store(&mounts_running, false);
+		return NULL;
+	}
+
+	fd = *((int *)arg);
+	free(arg);
+
+	handle_mounts(fd);
+	atomic_store(&mounts_running, false);
+
+	return NULL;
+}
+
+/*
+ * maybe_start_mounts_thread - start a detached mounts thread when needed.
+ * @fd: /proc/mounts file descriptor.
+ * Returns nothing.
+ */
+static void maybe_start_mounts_thread(int fd)
+{
+	int rc;
+	bool expected = false;
+	int *arg;
+
+	// Make sure one is not runnning since it is detached
+	if (!atomic_compare_exchange_strong(&mounts_running, &expected, true))
+		return;
+
+	arg = malloc(sizeof(int));
+	if (arg == NULL) {
+		msg(LOG_ERR, "Failed allocating mount thread arg");
+		atomic_store(&mounts_running, false);
+		return;
+	}
+	*arg = fd;
+
+	pthread_t tid;
+	pthread_attr_t attr;
+	pthread_attr_init(&attr);
+	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
+
+	rc = pthread_create(&tid, &attr,
+		handle_mounts_thread_main, arg);
+	if (rc) {
+		msg(LOG_ERR, "Failed starting mount thread (%s)",
+				strerror(rc));
+		atomic_store(&mounts_running, false);
+		free(arg);
+	}
+	pthread_attr_destroy(&attr);
+}
+
 
 static void usage(void)
 {
 	fprintf(stderr,
 		"Usage: fapolicyd [--debug|--debug-deny] [--permissive] "
-		"[--no-details]\n");
+		"[--no-details] [--version]\n");
 	exit(1);
 }
 
@@ -427,34 +738,38 @@ static void usage(void)
 static struct mallinfo2 last_mi;
 static void memory_use_report(FILE *f)
 {
-        struct mallinfo2 mi = mallinfo2();
+	struct mallinfo2 mi = mallinfo2();
 
-        fprintf(f, "glibc arena (total memory) is: %zu KiB, was: %zu KiB\n",
-                        (size_t)mi.arena/1024, (size_t)last_mi.arena/1024);
-        fprintf(f, "glibc uordblks (in use memory) is: %zu KiB, was: %zu KiB\n",
-                        (size_t)mi.uordblks/1024,(size_t)last_mi.uordblks/1024);
-        fprintf(f,"glibc fordblks (total free space) is: %zu KiB, was: %zu KiB\n",
-                        (size_t)mi.fordblks/1024,(size_t)last_mi.fordblks/1024);
+	fprintf(f, "glibc arena (total memory) is: %zu KiB, was: %zu KiB\n",
+			(size_t)mi.arena/1024, (size_t)last_mi.arena/1024);
+	fprintf(f, "glibc uordblks (in use memory) is: %zu KiB, was: %zu KiB\n",
+			(size_t)mi.uordblks/1024,(size_t)last_mi.uordblks/1024);
+	fprintf(f,"glibc fordblks (total free space) is: %zu KiB, was: %zu KiB\n",
+			(size_t)mi.fordblks/1024,(size_t)last_mi.fordblks/1024);
 
-        memcpy(&last_mi, &mi, sizeof(struct mallinfo2));
+	memcpy(&last_mi, &mi, sizeof(struct mallinfo2));
 }
 
 static void close_memory_report(void)
 {
 	struct mallinfo2 mi = mallinfo2();
 
-        msg(LOG_DEBUG, "total memory: %zu KiB, was: %zu KiB",
-                        (size_t)mi.arena/1024, (size_t)last_mi.arena/1024);
-        msg(LOG_DEBUG, "in use memory: %zu KiB, was: %zu KiB",
-                        (size_t)mi.uordblks/1024,(size_t)last_mi.uordblks/1024);
-        msg(LOG_DEBUG,"total free memory: %zu KiB, was: %zu KiB",
-                        (size_t)mi.fordblks/1024,(size_t)last_mi.fordblks/1024);
+	msg(LOG_DEBUG, "total memory: %zu KiB, was: %zu KiB",
+			(size_t)mi.arena/1024, (size_t)last_mi.arena/1024);
+	msg(LOG_DEBUG, "in use memory: %zu KiB, was: %zu KiB",
+			(size_t)mi.uordblks/1024,(size_t)last_mi.uordblks/1024);
+	msg(LOG_DEBUG,"total free memory: %zu KiB, was: %zu KiB",
+			(size_t)mi.fordblks/1024,(size_t)last_mi.fordblks/1024);
 }
 #endif
 
 void do_stat_report(FILE *f, int shutdown)
 {
+	const char *ptr = lookup_integrity(config.integrity);
+
 	fprintf(f, "Permissive: %s\n", config.permissive ? "true" : "false");
+	fprintf(f, "Integrity: %s\n", ptr ? ptr : "unknown");
+	fprintf(f, "CPU cores: %ld\n", sysconf(_SC_NPROCESSORS_ONLN));
 	fprintf(f, "q_size: %u\n", config.q_size);
 	q_report(f);
 	decision_report(f);
@@ -462,10 +777,20 @@ void do_stat_report(FILE *f, int shutdown)
 #ifdef HAVE_MALLINFO2
 	memory_use_report(f);
 #endif
+	if (!shutdown)
+		do_cache_reports(f);
+
+	// Report mounts under fanotify watch
+	if (m) {
+		const char *path = mlist_first(m);
+		while (path) {
+			fprintf(f, "watching mount: %s\n", path);
+			path = mlist_next(m);
+		}
+	}
+
 	if (shutdown)
 		fputs("\n", f);
-	else
-		do_cache_reports(f);
 }
 
 int already_running(void)
@@ -536,15 +861,20 @@ int main(int argc, const char *argv[])
 
 	setlocale(LC_TIME, "");
 
-	if (argc > 1 && strcmp(argv[1], "--help") == 0)
-		usage();
+	for (int i=1; i < argc; i++) {
+		if (strcmp(argv[i], "--help") == 0)
+			usage();
+		else if (strcmp(argv[i], "--version") == 0) {
+			printf("fapolicyd %s\n", VERSION);
+			return 0;
+		}
+	}
 	set_message_mode(MSG_STDERR, debug_mode);
 	if (load_daemon_config(&config)) {
 		free_daemon_config(&config);
 		msg(LOG_ERR, "Exiting due to bad configuration");
 		return 1;
-	}
-	permissive = config.permissive;
+        }
 
 	// set the debug flags
 	for (int i=1; i < argc; i++) {
@@ -559,7 +889,7 @@ int main(int argc, const char *argv[])
 	// process remaining flags
 	for (int i=1; i < argc; i++) {
 		if (strcmp(argv[i], "--permissive") == 0) {
-			permissive = 1;
+			config.permissive = 1;
 		} else if (strcmp(argv[i], "--boost") == 0) {
 			i++;
 			msg(LOG_ERR, "boost value on the command line is"
@@ -684,6 +1014,7 @@ int main(int argc, const char *argv[])
 	}
 
 	// Setup filesystem to watch list
+	init_ignore_mounts(config.ignore_mounts);
 	init_fs_list(config.watch_fs);
 
 	// Write the pid file for the init system
@@ -722,7 +1053,8 @@ int main(int argc, const char *argv[])
 	if (init_database(&config)) {
 		destroy_event_system();
 		destroy_rules();
-		destroy_fs_list();
+		destroy_fs_list(&filesystems);
+		destroy_fs_list(&ignored_mounts);
 		free_daemon_config(&config);
 		unlink(pidfile);
 		exit(1);
@@ -744,7 +1076,7 @@ int main(int argc, const char *argv[])
 		if (hup) {
 			hup = false;
 			msg(LOG_DEBUG, "Got SIGHUP");
-			reconfigure();
+			maybe_start_reconfigure_thread();
 		}
 		rc = poll(pfd, 2, -1);
 
@@ -756,9 +1088,10 @@ int main(int argc, const char *argv[])
 				continue;
 			else {
 				stop = true;
+				nudge_queue();
 				close_database();
 				msg(LOG_ERR, "Poll error (%s)\n",
-						strerror(errno));
+					strerror(errno));
 				exit(1);
 			}
 		} else if (rc > 0) {
@@ -767,7 +1100,7 @@ int main(int argc, const char *argv[])
 			}
 			if (pfd[0].revents & POLLPRI) {
 				msg(LOG_DEBUG, "Mount change detected");
-				handle_mounts(pfd[0].fd);
+				maybe_start_mounts_thread(pfd[0].fd);
 			}
 
 			// This will always need to be here as long as we
@@ -806,7 +1139,8 @@ int main(int argc, const char *argv[])
 	free(m);
 	destroy_event_system(); // clears lru caches
 	destroy_rules();
-	destroy_fs_list();
+	destroy_fs_list(&filesystems);
+	destroy_fs_list(&ignored_mounts);
 	free_daemon_config(&config);
 
 	return 0;
