@@ -28,6 +28,7 @@
 #include "daemon-config.h"
 #include "message.h"
 #include "file.h"
+#include "database.h"
 
 #include <stdio.h>
 #include <errno.h>
@@ -84,6 +85,8 @@ static int do_stat_report_parser(const struct nv_pair *nv, int line,
 		conf_t *config);
 static int watch_fs_parser(const struct nv_pair *nv, int line,
 		conf_t *config);
+static int ignore_mounts_parser(const struct nv_pair *nv, int line,
+		conf_t *config);
 static int trust_parser(const struct nv_pair *nv, int line,
 			   conf_t *config);
 static int integrity_parser(const struct nv_pair *nv, int line,
@@ -110,6 +113,7 @@ static const struct kw_pair keywords[] =
   {"obj_cache_size",	obj_cache_size_parser },
   {"do_stat_report",	do_stat_report_parser },
   {"watch_fs",		watch_fs_parser },
+  {"ignore_mounts",	ignore_mounts_parser },
   {"trust",		trust_parser },
   {"integrity",		integrity_parser },
   {"syslog_format",	syslog_format_parser },
@@ -132,9 +136,11 @@ static void clear_daemon_config(conf_t *config)
 	config->do_stat_report = 1;
 	config->detailed_report = 1;
 	config->db_max_size = 100;
+	config->do_audit_db_sizing = false;
 	config->subj_cache_size = 4099;
 	config->obj_cache_size = 8191;
 	config->watch_fs = strdup("ext4,xfs,tmpfs");
+	config->ignore_mounts = NULL;
 #ifdef USE_RPM
 	config->trust = strdup("rpmdb,file");
 #else
@@ -252,11 +258,16 @@ static char *get_line(FILE *f, char *buf, unsigned size, int *lineno,
 			too_long = 0;
 			*lineno = *lineno + 1;
 		} else {
-			// If a line is too long skip it.
-			// Only output 1 warning
-			if (!too_long)
+			if (!too_long) {
+				if (feof(f)) {
+					// last line without trailing newline
+					return buf;
+				}
+				// If a line is too long skip it.
+				// Only output 1 warning
 				msg(LOG_ERR, "Skipping line %d in %s: too long",
 					*lineno, file);
+			}
 			too_long = 1;
 		}
 	}
@@ -349,6 +360,7 @@ static const struct kw_pair *kw_lookup(const char *val)
 void free_daemon_config(conf_t *config)
 {
 	free((void*)config->watch_fs);
+	free((void*)config->ignore_mounts);
 	free((void*)config->trust);
 	free((void*)config->syslog_format);
 }
@@ -484,6 +496,14 @@ static int detailed_report_parser(const struct nv_pair *nv, int line,
 static int db_max_size_parser(const struct nv_pair *nv, int line,
 		conf_t *config)
 {
+	// "auto" triggers utilisation‑based sizing, anything else
+	// remains the legacy integer in MiB.
+	if (strcmp(nv->value, "auto") == 0) {
+		config->do_audit_db_sizing = true;
+		config->db_max_size = get_default_db_max_size();
+		return 0;
+	}
+
 	return unsigned_int_parser(&(config->db_max_size), nv->value, line);
 }
 
@@ -533,6 +553,24 @@ static int watch_fs_parser(const struct nv_pair *nv, int line,
 	return 1;
 }
 
+/*
+ * ignore_mounts_parser - store ignore_mounts configuration setting.
+ * @nv: name/value pair describing the option.
+ * @line: line number where the option was found.
+ * @config: configuration structure to update.
+ * Returns 0 on success and 1 when memory cannot be allocated.
+ */
+static int ignore_mounts_parser(const struct nv_pair *nv, int line,
+		conf_t *config)
+{
+	free((void *)config->ignore_mounts);
+	config->ignore_mounts = strdup(nv->value);
+	if (config->ignore_mounts)
+		return 0;
+	msg(LOG_ERR, "Could not store value line %d", line);
+	return 1;
+}
+
 static int report_interval_parser(const struct nv_pair *nv, int line,
         conf_t *config)
 {
@@ -571,8 +609,9 @@ static int integrity_parser(const struct nv_pair *nv, int line,
 				int fd = open("/bin/sh", O_RDONLY);
 				if (fd >= 0) {
 					char sha[65];
+					file_hash_alg_t alg;
 
-					int rc = get_ima_hash(fd, sha);
+					int rc = get_ima_hash(fd, &alg, sha);
 					close(fd);
 					if (rc == 0) {
 						msg(LOG_ERR,
@@ -592,6 +631,13 @@ static int integrity_parser(const struct nv_pair *nv, int line,
 	return 1;
 }
 
+const char *lookup_integrity(unsigned value)
+{
+	if (value > 3)
+		return NULL;
+
+	return integrity_schemes[value].name;
+}
 
 static int syslog_format_parser(const struct nv_pair *nv, int line,
 		conf_t *config)
