@@ -32,6 +32,7 @@
 #include <stdio.h>
 #include <stdbool.h>
 #include <time.h>
+#include <errno.h>
 
 #include "event.h"
 #include "database.h"
@@ -39,12 +40,16 @@
 #include "lru.h"
 #include "message.h"
 #include "policy.h"
+#include "rules.h"
+#include "process.h"
 
 #define ALL_EVENTS (FAN_ALL_EVENTS|FAN_OPEN_PERM|FAN_ACCESS_PERM| \
 	FAN_OPEN_EXEC_PERM)
 
 static Queue *subj_cache = NULL;
 static Queue *obj_cache = NULL;
+static bool obj_cache_warned = false;
+static unsigned int early_subj_cache_evictions = 0;
 
 atomic_bool needs_flush = false;
 
@@ -61,9 +66,7 @@ atomic_bool needs_flush = false;
  */
 static void subject_evict_warn(s_array *s)
 {
-	int pid = -1;
-	if (s && s->info)
-		pid = s->info->pid;
+	int warn = 0;
 	if (s && s->info && s->info->state < STATE_FULL) {
 		/*
 		 * Normal interpreter re-exec replaces the process image
@@ -72,12 +75,93 @@ static void subject_evict_warn(s_array *s)
 		 * Suppress the suggestion to grow the cache.
 		 */
 		if (!((s->info->state == STATE_REOPEN) &&
-		      (s->info->elf_info & (HAS_SHEBANG|TEXT_SCRIPT))) )
-			msg(LOG_WARNING,
-			    "pid %d in state %d (%s) is being evicted from the "
-			    "subject cache before pattern detection completes: "
-			    "increase subj_cache_size",
-			    pid, s->info->state, s->info->path1);
+		      (s->info->elf_info & (HAS_SHEBANG|TEXT_SCRIPT))) ) {
+			warn = 1;
+			early_subj_cache_evictions++;
+		}
+	}
+
+	if (early_subj_cache_evictions > 5)
+		return;
+
+	if (warn) {
+		msg(LOG_WARNING,
+		    "pid %d in state %d (%s) is being evicted from the "
+		    "subject cache before pattern detection completes: "
+		    "increase subj_cache_size",
+		    s->info->pid, s->info->state, s->info->path1);
+	}
+}
+
+/*
+ * obj_evict_warn - check object cache eviction ratios
+ *
+ * Opportunistically check eviction ratios during evictions and warn the
+ * administrator when thresholds indicate the object cache is too small.
+ * Checks occur no more than once every 16 evictions and only one runtime
+ * warning is emitted.
+ *
+ * It uses 2 ratios to decide if we need to issue a warning:
+ *
+ * E_over_M = evictions / misses
+ *   - Measures how often a miss requires throwing out an existing object.
+ *   - High values mean the cache is not just missing, but actively churning,
+ *     which points to either capacity pressure or poor distribution.
+ *
+ * E_over_Q = evictions / total lookups
+ *   - Measures the overall fraction of requests that cause an eviction.
+ *   - This gives a user-facing view of churn: how much of the workload is
+ *     paying the eviction penalty out of all operations.
+ *
+ * Together, these ratios let us distinguish "expected misses" from
+ * "pathological evictions" and trigger a resize warning only when the cache
+ * is turning over too aggressively for its occupancy level.
+ */
+__attribute__((cold))
+static void obj_evict_warn(void *unused)
+{
+	unsigned long evicts, miss, hit, lookups, e_over_m, e_over_q;
+	unsigned int occ, thr_m = 0, thr_q = 0;
+
+	if (obj_cache_warned)
+		return;
+
+	if (obj_cache->evictions & 0xF)
+		return;
+
+	evicts = obj_cache->evictions + 1;
+	miss = obj_cache->misses + 1;
+	hit = obj_cache->hits;
+	lookups = hit + miss;
+	occ = (obj_cache->count * 100) / obj_cache->total;
+	e_over_m = (evicts * 100) / miss;
+	e_over_q = (evicts * 100) / (lookups ? lookups : 1);
+
+	if (occ >= 85) {
+		// Near-full tables churn; above these levels growth is
+		// usually cheaper than misses.
+		thr_m = 80;
+		thr_q = 35;
+	} else if (occ >= 75) {
+		// Some churn is expected; beyond this you’re throwing away
+		// too much reuse.
+		thr_m = 55;
+		thr_q = 20;
+	} else if (occ >= 60) {
+		// At this level evictions should be infrequent; higher means
+		// collisions/skew or underprovisioning.
+		thr_m = 35;
+		thr_q = 12;
+	} else
+		return;
+
+	if (e_over_m > thr_m || e_over_q > thr_q) {
+		msg(LOG_WARNING,
+		    "object cache eviction ratios high (occupancy: %u%%, "
+		    "evict/miss=%lu%%, evict/lookups=%lu%%): "
+		    "increase obj_cache_size",
+		    occ, e_over_m, e_over_q);
+		obj_cache_warned = true;
 	}
 }
 
@@ -98,9 +182,12 @@ int init_event_system(const conf_t *config)
 
 	obj_cache = init_lru(config->obj_cache_size,
 				(void (*)(void *))object_clear, "Object",
-				NULL);
-	if (!obj_cache)
+				obj_evict_warn);
+	if (!obj_cache) {
+		destroy_lru(subj_cache);
+		subj_cache = NULL;
 		return 1;
+	}
 
 	return 0;
 }
@@ -113,11 +200,12 @@ static int flush_cache(void)
 	const unsigned int size = obj_cache->total;
 
 	msg(LOG_DEBUG, "Flushing object cache");
+	obj_cache->evict_cb = NULL;
 	destroy_lru(obj_cache);
 
 	obj_cache = init_lru(size,
 				(void (*)(void *))object_clear, "Object",
-				NULL);
+				obj_evict_warn);
 	if (!obj_cache)
 		return 1;
 
@@ -131,10 +219,28 @@ void destroy_event_system(void)
 	/* We're intentionally clearing the caches; disable warnings */
 	if (subj_cache)
 		subj_cache->evict_cb = NULL;
+	if (early_subj_cache_evictions)
+		msg(LOG_WARNING,
+		   "Processes are being evicted from the subject cache before "
+		   "pattern detection completes: increase subj_cache_size "
+		   "(total early evictions: %u)", early_subj_cache_evictions);
+	if (obj_cache)
+		obj_cache->evict_cb = NULL;
+	if (obj_cache_warned)
+		msg(LOG_WARNING,
+		  "object cache eviction ratios high: increase obj_cache_size");
 	destroy_lru(subj_cache);
 	destroy_lru(obj_cache);
 }
 
+static inline void reset_subject_attributes(s_array *s)
+{
+	subject_reset(s, EXE);
+	subject_reset(s, COMM);
+	subject_reset(s, EXE_TYPE);
+	subject_reset(s, SUBJ_TRUST);
+}
+
 // Return 0 on success and 1 on error
 int new_event(const struct fanotify_event_metadata *m, event_t *e)
 {
@@ -146,9 +252,9 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 	struct proc_info *pinfo;
 	struct file_info *finfo;
 
-	if (needs_flush) {
+	if (atomic_exchange_explicit(&needs_flush, false,
+				     memory_order_acq_rel)) {
 		flush_cache();
-		needs_flush = false;
 	}
 
 	// Transfer things from fanotify structs to ours
@@ -178,14 +284,15 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 		// because it was collected on perm = execute.
 		if ((s->info->state == STATE_COLLECTING) &&
 			(e->type & FAN_OPEN_PERM) && !rc) {
-			skip_path = 1;
-			s->info->state = STATE_REOPEN;
-
 			// special branch after ld_so exec
 			// next opens will go fall trough
 			if (s->info->path1 &&
 				(strcmp(s->info->path1, SYSTEM_LD_SO) == 0))
 				s->info->state = STATE_DEFAULT_REOPEN;
+			else {
+				skip_path = 1;
+				s->info->state = STATE_REOPEN;
+			}
 		}
 
 		// If not same proc or we detect execution, evict
@@ -202,10 +309,7 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 				skip_path = 1;
 			}
 			evict = 0;
-			subject_reset(s, EXE);
-			subject_reset(s, COMM);
-			subject_reset(s, EXE_TYPE);
-			subject_reset(s, SUBJ_TRUST);
+			reset_subject_attributes(s);
 		}
 		// Static has to sequence through a state machine to get to
 		// the point where we can do a full subject reset. Still
@@ -252,7 +356,7 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 		e->s = malloc(sizeof(s_array));
 		subject_create(e->s);
 		subj.type = PID;
-		subj.val = e->pid;
+		subj.pid = e->pid;
 		subject_add(e->s, &subj);
 
 		// give custody of the list to the cache
@@ -274,8 +378,16 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 	// get file fingerprint
 	rc = 1;
 	finfo = stat_file_entry(m->fd);
-	if (finfo == NULL)
+	if (finfo == NULL) {
+		/* On stat_file_entry failure, evict the subject to avoid
+		 * leaving an incomplete subject cached, which could
+		 * confuse later lookups and pattern matching. */
+		if (evict) {
+			lru_evict(subj_cache, key);
+			e->s = NULL;
+		}
 		return 1;
+	}
 
 	// Just using inodes don't give a good key. It needs
 	// conditioning to use more slots in the cache.
@@ -328,20 +440,83 @@ int new_event(const struct fanotify_event_metadata *m, event_t *e)
 				// procfs is concerned. Reset things that could
 				// change based on the new process name.
 				pinfo->state = STATE_FULL;
-				subject_reset(e->s, EXE);
-				subject_reset(e->s, COMM);
-				subject_reset(e->s, EXE_TYPE);
-				subject_reset(e->s, SUBJ_TRUST);
+				reset_subject_attributes(s);
 			}
-		} 
+		}
 	}
 	return 0;
 }
 
 /*
- * This function will search the list for a nv pair of the right type.
- * If not found, it will create the type and return it.
+ * fetch_proc_status - populate subject cache entries using /proc status
+ * @e: event whose subject cache should be filled
+ * @t: subject attribute type requested by the caller
+ *
+ * The function gathers all configured fields from /proc/<pid>/status for the
+ * process associated with @e.  Each successfully read attribute is added to
+ * the subject cache so subsequent lookups do not need to touch procfs
+ * again.
+ *
+ * Return: pointer to the requested attribute on success, NULL otherwise.
  */
+subject_attr_t *fetch_proc_status(event_t *e, subject_type_t t)
+{
+	unsigned int mask = rules_get_proc_status_mask();
+	mask |= policy_get_syslog_proc_status_mask();
+	struct proc_status_info info = {
+		.ppid = -1,
+		.uid = NULL,
+		.groups = NULL,
+		.comm = NULL
+	};
+
+	if (read_proc_status(e->pid, mask, &info) != 0)
+		return NULL;
+
+	// Cache everything - sets and comm are malloc'ed. Transfer ownership.
+	// Not checking return of subject_add. Caller needs to check for NULL.
+	if (mask & PROC_STAT_PPID) {
+		subject_attr_t sub;
+		sub.type = PPID;
+		sub.pid = info.ppid;
+		subject_add(e->s, &sub);
+	}
+	if (mask & PROC_STAT_UID) {
+		subject_attr_t sub;
+		sub.type = UID;
+		sub.set = info.uid;
+		subject_add(e->s, &sub);
+	}
+	if (mask & PROC_STAT_GID) {
+	    subject_attr_t sub;
+	    sub.type = GID;
+	    sub.set = info.groups;
+	    subject_add(e->s, &sub);
+	}
+	if (mask & PROC_STAT_COMM) {
+		subject_attr_t sub;
+		sub.type = COMM;
+		sub.str = info.comm;
+		subject_add(e->s, &sub);
+	}
+
+	//return the subject entry
+	return subject_access(e->s, t);
+}
+
+/*
+ * get_subj_attr - return a subject attribute, creating it on demand
+ * @e: event describing the subject whose attribute is needed
+ * @t: subject attribute identifier
+ *
+ * The function first looks for @t in the subject cache.  When missing, it
+ * performs the necessary lookup and stores the result for reuse.  Some
+ * attributes are retrieved directly, while UID/GID and credential data
+ * are collected in bulk via fetch_proc_status().
+ *
+ * Return: pointer to the requested attribute, or NULL if acquisition fails.
+ */
+__attribute__((hot))
 subject_attr_t *get_subj_attr(event_t *e, subject_type_t t)
 {
 	subject_attr_t subj;
@@ -352,50 +527,55 @@ subject_attr_t *get_subj_attr(event_t *e, subject_type_t t)
 	if (sn)
 		return sn;
 
-	// One not on the list, look it up and make one
+	// The desired attribute is not on the list, look it up and cache it
 	subj.type = t;
 	subj.str = NULL;
 	switch (t) {
 		case AUID:
-			subj.val = get_program_auid_from_pid(e->pid);
-			break;
-		case UID:
-			subj.val = get_program_uid_from_pid(e->pid);
-			break;
-		case SESSIONID:
-			subj.val = get_program_sessionid_from_pid(e->pid);
-			break;
-		case PID:
-			subj.val = e->pid;
+			subj.uval = get_program_auid_from_pid(e->pid);
 			break;
 		case PPID:
-			subj.val = get_program_ppid_from_pid(e->pid);
-			break;
+		case UID:
 		case GID:
-			subj.set = get_gid_set_from_pid(e->pid);
+		case COMM:
+			/*
+			 * UID/GID credentials may differ between the real,
+			 * effective, saved, and filesystem slots.  Cache all
+			 * but saved so the rule engine can evaluate all
+			 * possible identities during matching.
+			 */
+			return fetch_proc_status(e, t);
 			break;
-		case COMM: {
-			char buf[21], *ptr;
-			ptr = get_comm_from_pid(e->pid,	sizeof(buf), buf);
-			if (ptr)
-				subj.str = strdup(buf);
-			else
-				subj.str = strdup("?");
-			}
+		case SESSIONID:
+			subj.uval = (unsigned int)
+					get_program_sessionid_from_pid(e->pid);
+			break;
+		case PID:
+			subj.pid = e->pid;
 			break;
 		// If these 2 ever get separated, update subject_add
 		// and subject_access in subject.c
 		case EXE:
 		case EXE_DIR: {
 			char buf[PATH_MAX+1], *ptr;
-			ptr = get_program_from_pid(e->pid,
-						sizeof(buf), buf);
-			if (ptr)
+
+			errno = 0;
+			ptr = get_program_from_pid(e->pid, sizeof(buf), buf);
+			if (errno == ENOENT) {
+				/* kworkers have no exe entry, use comm */
+				sn = subject_access(s, COMM);
+				if (!sn)
+					sn = fetch_proc_status(e, COMM);
+				if (sn)
+					subj.str = strdup(sn->str);
+				else
+					subj.str = strdup("?");
+			} else if (ptr)
 				subj.str = strdup(buf);
 			else
 				subj.str = strdup("?");
-			}
-			break;
+		}
+		break;
 		case EXE_TYPE: {
 			char buf[128], *ptr;
 			ptr = get_type_from_pid(e->pid, sizeof(buf), buf);
@@ -408,16 +588,17 @@ subject_attr_t *get_subj_attr(event_t *e, subject_type_t t)
 		case SUBJ_TRUST: {
 			subject_attr_t *exe = get_subj_attr(e, EXE);
 
-			subj.val = 0;
+			subj.uval = 0;
 			if (exe) {
 				if (exe->str) {
-					int res = check_trust_database(exe->str, NULL, 0);
+					int res = check_trust_database(exe->str,
+								       NULL, 0);
 
 					// ignore -1
 					if (res == 1)
-						subj.val = 1;
+						subj.uval = 1;
 					else
-						subj.val = 0;
+						subj.uval = 0;
 				}
 			}
 			}
@@ -431,11 +612,9 @@ subject_attr_t *get_subj_attr(event_t *e, subject_type_t t)
 		return sn;
 	}
 
-	// free .str/set only when it was really used
+	// free .str only when it was really used
 	// otherwise invalid free is possible
-	if (t == GID)
-		destroy_attr_set(subj.set);
-	else if (t >= COMM)
+	if (t >= COMM)
 		free(subj.str);
 	return NULL;
 }
@@ -445,6 +624,7 @@ subject_attr_t *get_subj_attr(event_t *e, subject_type_t t)
  * This function will search the list for a nv pair of the right type.
  * If not found, it will create the type and return it.
  */
+__attribute__((hot))
 object_attr_t *get_obj_attr(event_t *e, object_type_t t)
 {
 	char buf[PATH_MAX+1], *ptr;
@@ -495,14 +675,36 @@ object_attr_t *get_obj_attr(event_t *e, object_type_t t)
 				obj.o = strdup("?");
 			}
 			break;
-		case SHA256HASH:
-			obj.o = get_hash_from_fd2(e->fd, o->info->size, 1);
-			break;
+		case FILE_HASH: {
+			file_hash_alg_t alg = FILE_HASH_ALG_SHA256;
+
+			if (o->info) {
+				if (o->info->digest_alg != FILE_HASH_ALG_NONE)
+					alg = o->info->digest_alg;
+
+				if (o->info->digest[0]) {
+					obj.o = strdup(o->info->digest);
+					break;
+				}
+
+				obj.o = get_hash_from_fd2(e->fd,
+							  o->info->size, alg);
+				if (obj.o) {
+					file_info_cache_digest(o->info, alg);
+					strncpy(o->info->digest, obj.o,
+						FILE_DIGEST_STRING_MAX-1);
+					o->info->digest[FILE_DIGEST_STRING_MAX-1] = 0;
+				} else
+					file_info_reset_digest(o->info);
+			}
+		}
+		break;
 		case OBJ_TRUST: {
 			object_attr_t *path =  get_obj_attr(e, PATH);
 
 			if (path && path->o) {
-				int res = check_trust_database(path->o, o->info, e->fd);
+				int res = check_trust_database(path->o,
+							       o->info, e->fd);
 
 				// ignore -1
 				if (res == 1)
@@ -643,6 +845,8 @@ void run_usage_report(const conf_t *config, FILE *f)
 void do_cache_reports(FILE *f)
 {
 	print_queue_stats(f, subj_cache);
+	fprintf(f, "Early subject cache evictions: %u\n",
+		early_subj_cache_evictions);
 	print_queue_stats(f, obj_cache);
 }
 
