@@ -26,6 +26,7 @@
 #ifndef EVENT_HEADER
 #define EVENT_HEADER
 
+#include "config.h"
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/fanotify.h>
@@ -45,8 +46,8 @@ typedef struct ev {
 int init_event_system(const conf_t *config);
 void destroy_event_system(void);
 int new_event(const struct fanotify_event_metadata *m, event_t *e);
-subject_attr_t *get_subj_attr(event_t *e, subject_type_t t);
-object_attr_t *get_obj_attr(event_t *e, object_type_t t);
+__attribute__((hot)) subject_attr_t *get_subj_attr(event_t *e, subject_type_t t);
+__attribute__((hot)) object_attr_t *get_obj_attr(event_t *e, object_type_t t);
 void run_usage_report(const conf_t *config, FILE *f);
 void do_cache_reports(FILE *f);
 
