@@ -75,6 +75,8 @@ void filter_set_trace(FILE *stream)
 
 static filter_t *filter_create_obj(void);
 static void filter_destroy_obj(filter_t *_filter);
+static int stack_push_vars(stack_t *_stack, stack_item_t *buf, int *sp,
+			    int _level, int _offset, filter_t *_filter);
 
 /*
  * filter_init - initialize module and global filter tree
@@ -237,8 +239,10 @@ static void stack_pop_all_reset(stack_t *_stack, int *sp)
  * filter_check - compare path against loaded filters
  * @_path: full path of file to test
  * Returns FILTER_ALLOW if file should be kept, FILTER_DENY if it should be
- * dropped, or FILTER_ERR_DEPTH if MAX_FILTER_DEPTH is exceeded.
+ * dropped, or FILTER_ERR_DEPTH if MAX_FILTER_DEPTH is exceeded (treated the
+ * same as a deny by callers to keep processing other paths).
  */
+__attribute__((hot))
 filter_rc_t filter_check(const char *_path)
 {
 	if (_path == NULL) {
@@ -250,7 +254,11 @@ filter_rc_t filter_check(const char *_path)
 	size_t path_len = strlen(_path);
 	char *path = alloca(path_len + 1);
 	strcpy(path, _path);
-	if (strstr(path, "/../"))
+	/* Reject paths with parent directory references */
+	if ((path[0] == '.' && path[1] == '.' &&
+		(path[2] == '/' || path[2] == '\0')) ||
+		strstr(path, "/../") != NULL ||
+		    (path_len >= 3 && strcmp(path + path_len - 3, "/..") == 0))
 		return FILTER_DENY;
 	/* offset tracks how much of the path has already matched */
 	size_t offset = 0;
@@ -340,9 +348,8 @@ filter_rc_t filter_check(const char *_path)
 				if (count && *(filter_old_lim+1) == '\0')
 					*(path_old_lim+1) = tmp;
 
-				if (matched) {
-					offset = path_old_lim - path+offset;
-				}
+				if (matched)
+					offset = (path_old_lim - path) + offset;
 			} else {
 				// match normal path or just specific part of it
 				matched = !strncmp(path+offset, filter->path,
@@ -357,11 +364,18 @@ filter_rc_t filter_check(const char *_path)
 
 				// if matched we need ot push descendants
 				// to the stack
-				list_item_t *item = list_get_first(&filter->list);
+				list_item_t *item=list_get_first(&filter->list);
 
 				// if there are no descendants and it is
 				// a wildcard then it's a match
 				if (item == NULL && is_wildcard) {
+					const char *rule = (filter->path &&
+							*filter->path) ?
+							filter->path : "/";
+					FILTER_TRACE("%s %s %s\n",
+						     filter->type == ADD ?
+						     "allow" : "deny",
+						     rule, "match");
 					// if '+' ret 1 and if '-' ret 0
 					res = filter->type == ADD ?
 						FILTER_ALLOW : FILTER_DENY;
@@ -371,6 +385,13 @@ filter_rc_t filter_check(const char *_path)
 				// no descendants, and already compared
 				// whole path string so its a match
 				if (item == NULL && path_len == offset) {
+					const char *rule = (filter->path &&
+							*filter->path) ?
+							filter->path : "/";
+					FILTER_TRACE("%s %s %s\n",
+							filter->type == ADD ?
+							"allow" : "deny",
+							rule, "match");
 					// if '+' ret 1 and if '-' ret 0
 					res = filter->type == ADD ?
 						FILTER_ALLOW : FILTER_DENY;
@@ -390,9 +411,7 @@ filter_rc_t filter_check(const char *_path)
 						goto end;
 					}
 				}
-
 			}
-
 		}
 
 		if (filter->type != NONE) {
@@ -444,6 +463,58 @@ end:
 	return res;
 }
 
+/*
+ * filter_prune_list - Remove list entries that do not pass the filter.
+ * @list: List of paths to be checked.
+ * @path: Optional configuration file path, defaults to FILTER_FILE.
+ *
+ * Initializes the filter module, loads the configuration, and walks the
+ * supplied list. Any entry that is not allowed by the filter is removed.
+ * Returns 0 on success and 1 if initialization, loading, or evaluation fails.
+ */
+int filter_prune_list(list_t *list, const char *path)
+{
+	if (list == NULL)
+		return 1;
+
+	if (filter_init())
+		return 1;
+	if (filter_load_file(path)) {
+		filter_destroy();
+	return 1;
+	}
+
+	list_item_t *lptr = list->first, *prev = NULL;
+
+	while (lptr) {
+		list_item_t *next = lptr->next;
+		filter_rc_t res = filter_check(lptr->index);
+		if (res == FILTER_ALLOW) {
+			prev = lptr;
+			lptr = next;
+			continue;
+		}
+
+		if (res == FILTER_ERR_DEPTH)
+			msg(LOG_WARNING,
+			    "filter nesting exceeds MAX_FILTER_DEPTH for %s; excluding",
+			    (char *)lptr->index);
+
+		if (prev)
+			prev->next = lptr->next;
+		else
+			list->first = lptr->next;
+		if (!lptr->next)
+			list->last = prev;
+		list_destroy_item(&lptr);
+		--list->count;
+		lptr = next;
+	}
+
+	filter_destroy();
+	return 0;
+}
+
 /*
  * filter_load_file - load filter configuration and build tree
  * @path: optional configuration file path, defaults to FILTER_FILE
@@ -454,6 +525,7 @@ int filter_load_file(const char *path)
 	int res = 0;
 	FILE *stream;
 
+	msg(LOG_DEBUG, "Loading filter");
 	if (path == NULL) {
 		stream = fopen(OLD_FILTER_FILE, "r");
 
