@@ -63,6 +63,8 @@ static list_item_t * create_item(const char *index, const char *data)
 int list_prepend(list_t *list, const char *index, const char *data)
 {
 	list_item_t *item = create_item(index, data);
+	if (item == NULL)
+	    return 1;
 
 	item->next = list->first;
 	list->first = item;
@@ -75,6 +77,9 @@ int list_append(list_t *list, const char *index, const char *data)
 {
 	list_item_t *item = create_item(index, data);
 
+	if (!item)
+		return 1;
+
 	if (list->first) {
 		list->last->next = item;
 		list->last = item;
@@ -113,7 +118,7 @@ void list_empty(list_t *list)
 
 
 // Return 1 if the list contains the string, 0 otherwise
-int list_contains(list_t *list, const char *str)
+int list_contains(const list_t *list, const char *str)
 {
 	for (list_item_t *lptr = list->first; lptr; lptr = lptr->next) {
 		if (!strcmp(str, lptr->index))
