@@ -44,7 +44,7 @@ int list_prepend(list_t *list, const char *index, const char *data);
 int list_append(list_t *list, const char *index, const char *data);
 void list_destroy_item(list_item_t **item);
 void list_empty(list_t *list);
-int list_contains(list_t *list, const char *str);
+int list_contains(const list_t *list, const char *str);
 int list_remove(list_t *list, const char *str);
 void list_merge(list_t *dest, list_t *src);
 
