@@ -29,6 +29,8 @@
 #include <sys/types.h>
 #include <sys/fanotify.h>
 #include <stdatomic.h>
+#include <semaphore.h>
+#include <time.h>
 #include "gcc-attributes.h"
 
 struct queue
@@ -36,9 +38,10 @@ struct queue
 	/* Ring buffer of fanotify events */
 	struct fanotify_event_metadata *events;
 	size_t num_entries;
-	size_t queue_head;
-	size_t queue_tail;
-	atomic_size_t queue_length;
+	atomic_uint q_next;
+	atomic_uint q_last;
+	atomic_uint queue_length;
+        sem_t sem;
 };
 
 /* Close Q. */
@@ -52,16 +55,24 @@ struct queue *q_open(size_t num_entries) __attribute_malloc__
 void q_report(FILE *f);
 
 /* Add DATA to tail of Q. Return 0 on success, -1 on error and set errno. */
-int q_append(struct queue *q, const struct fanotify_event_metadata *data);
+int q_enqueue(struct queue *q, const struct fanotify_event_metadata *data);
 
-/* Peek at head of Q, storing it into BUF of SIZE. Return 1 if an entry
- * exists, 0 if queue is empty. On error, return -1 and set errno. */
-int q_peek(const struct queue *q, struct fanotify_event_metadata *data);
+/* Remove one event from Q, storing it into DATA. Return 1 on success or 0 if
+ * the queue is empty. */
+int q_dequeue(struct queue *q, struct fanotify_event_metadata *data);
 
-/* Drop head of Q and return 0. On error, return -1 and set errno. */
-int q_drop_head(struct queue *q);
+/* Remove one event from Q, blocking until timeout. On success return 1. On
+ * timeout return 0 and set errno to ETIMEDOUT. */
+ int q_timed_dequeue(struct queue *q, struct fanotify_event_metadata *data,
+		     const struct timespec *ts);
+
+/* Wake up anyone waiting on the queue. */
+void q_shutdown(struct queue *q);
 
 /* Return the number of entries in Q. */
-static inline size_t q_queue_length(const struct queue *q) { return q->queue_length; }
+static inline size_t q_queue_length(const struct queue *q)
+{
+	return atomic_load_explicit(&q->queue_length, memory_order_relaxed);
+}
 
 #endif
