@@ -39,11 +39,11 @@
 #include "message.h"
 #include "file.h" // This seems wrong
 #include "database.h"
-
+#include "process.h"
 #include "subject-attr.h"
 #include "object-attr.h"
-
 #include "string-util.h"
+#include "gcc-attributes.h"
 
 //#define DEBUG
 #define UNUSED 0xFF
@@ -59,12 +59,21 @@
 #define PATTERN_LD_PRELOAD_STR "ld_preload"
 #define PATTERN_LD_PRELOAD_VAL 3
 
+static unsigned int proc_status_mask;
+
+static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
+	__wur;
+static int assign_object(lnode *n, int type, const char *ptr2, int lineno)
+	__wur;
+
 int rules_create(llist *l)
 {
 	l->head = NULL;
 	l->cur = NULL;
 	l->cnt = 0;
 
+	proc_status_mask = 0;
+
 	return  0;
 }
 
@@ -210,7 +219,7 @@ static bool is_subj_trusted(event_t *e)
 
 	if (!trusted)
 		return 0;
-	return trusted->val;
+	return trusted->uval;
 }
 
 
@@ -250,6 +259,20 @@ static char * parse_set_name(char * buf)
 
 #define GROUP_NAME_SIZE 64
 
+static const char *data_type_to_name(int type)
+{
+	switch (type) {
+		case STRING:
+			return "STRING";
+		case SIGNED:
+			return "SIGNED";
+		case UNSIGNED:
+			return "UNSIGNED";
+		default:
+			return "UNKNOWN";
+	}
+}
+
 static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 {
 	size_t index = 0;
@@ -260,6 +283,17 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 	sanity_check_node(n, "assign_subject - 1");
 	n->s[i].type = type;
 
+	// Opportunistically mark the fields that might be needed for
+	// rule evaluation so that we gather them all at once later.
+	if (type == UID)
+		proc_status_mask |= PROC_STAT_UID;
+	else if (type == PPID)
+		proc_status_mask |= PROC_STAT_PPID;
+	else if (type == GID)
+		proc_status_mask |= PROC_STAT_GID;
+	else if (type == COMM)
+		proc_status_mask |= PROC_STAT_COMM;
+
 	char *ptr, *saved, *tmp = strdup(ptr2);
 	if (tmp == NULL) {
 		msg(LOG_ERR, "memory allocation error in line %d",
@@ -298,19 +332,27 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 			goto free_and_error;
 		}
 
-		// numbers
-		if (type <= PPID && set->type != INT ) {
-			msg(LOG_ERR, "rules: line:%d: assign_subject: "
-				"cannot assign %%%s which has STRING type to INT",
-				lineno, defined_set);
-			goto free_and_error;
+		if (type <= PPID) {
+			int expected = (type == PID || type == PPID) ? SIGNED : UNSIGNED;
+			if (set->type != expected) {
+				msg(LOG_ERR, "rules: line:%d: assign_subject: "
+					"cannot assign %%%s which has %s type "
+					"to %s (%s expected)",
+					lineno, defined_set,
+					data_type_to_name(set->type),
+					subj_val_to_name(type, RULE_FMT_COLON),
+					data_type_to_name(expected));
+				goto free_and_error;
+			}
 		}
 
-		// strings
 		if (type >= COMM && set->type != STRING) {
 			msg(LOG_ERR, "rules: line:%d: assign_subject: "
-				"cannot assign %%%s which has STRING type to INT",
-				lineno, defined_set);
+				"cannot assign %%%s which has %s type to %s "
+				"(STRING expected)",
+				lineno, defined_set,
+				data_type_to_name(set->type),
+				subj_val_to_name(type, RULE_FMT_COLON));
 			goto free_and_error;
 		}
 
@@ -338,7 +380,10 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 	case GID:
 	case PID:
 	case PPID: {
-		if (add_attr_set(name, INT, &index)) {
+		int set_type = (n->s[i].type == PID ||
+				n->s[i].type == PPID) ? SIGNED : UNSIGNED;
+
+		if (add_attr_set(name, set_type, &index)) {
 			goto free_and_error;
 		}
 
@@ -348,17 +393,44 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 
 		ptr = strtok_r(tmp, ",", &saved);
 		while (ptr) {
-            // starts with a digit or minus sign
-			if (isdigit(*ptr) || *ptr == '-') {
+			ptr = fapolicyd_strtrim(ptr);
+			if (!ptr || *ptr == '\0') {
+				ptr = strtok_r(NULL, ",", &saved);
+				continue;
+			}
+			if (isdigit((unsigned char)*ptr) || *ptr == '-') {
 				errno = 0;
-				long val = strtol(ptr, NULL, 10);
-				if (errno) {
-					msg(LOG_ERR,
-					"Error converting val (%s) in line %d",
+				if (n->s[i].type == PID || n->s[i].type == PPID) {
+					long val = strtol(ptr, NULL, 10);
+					if (errno) {
+						msg(LOG_ERR,
+							"Error converting val (%s) in line %d",
 							ptr, lineno);
-					goto free_and_error;
-				} else if (append_int_attr_set(set, (int)val)) {
-					goto free_and_error;
+						goto free_and_error;
+					} else if (append_int_attr_set(set,
+							(int64_t)val)) {
+						goto free_and_error;
+					}
+				} else {
+					if (*ptr == '-') {
+						msg(LOG_ERR,
+							"rules: line:%d: assign_subject: "
+							"negative value %s not allowed for %s",
+							lineno, ptr,
+							subj_val_to_name(type,
+							    RULE_FMT_COLON));
+						goto free_and_error;
+					}
+					unsigned long val = strtoul(ptr, NULL, 10);
+					if (errno) {
+						msg(LOG_ERR,
+							"Error converting val (%s) in line %d",
+							ptr, lineno);
+						goto free_and_error;
+					} else if (append_int_attr_set(set,
+							(int64_t)val)) {
+						goto free_and_error;
+					}
 				}
 
 			// Support names for auid and uid entries
@@ -367,26 +439,26 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 				struct passwd *pw = getpwnam(ptr);
 				if (pw == NULL) {
 					msg(LOG_ERR, "user %s is unknown",
-							ptr);
+						ptr);
 					goto free_and_error;
 				}
-				int val = pw->pw_uid;
+				unsigned int val = pw->pw_uid;
 				endpwent();
 
-				if (append_int_attr_set(set, val))
+				if (append_int_attr_set(set, (int64_t)val))
 					goto free_and_error;
 
 			} else if (n->s[i].type == GID) {
 				struct group *gr = getgrnam(ptr);
 				if (gr == NULL) {
 					msg(LOG_ERR, "group %s is unknown",
-							ptr);
+						ptr);
 					goto free_and_error;
 				}
-				int val = gr->gr_gid;
+				unsigned int val = gr->gr_gid;
 				endgrent();
 
-				if (append_int_attr_set(set, val))
+				if (append_int_attr_set(set, (int64_t)val))
 					goto free_and_error;
 			}
 
@@ -407,13 +479,13 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 		}
 
 		if (strcmp(tmp,	PATTERN_LD_SO_STR) == 0) {
-			n->s[i].val = PATTERN_LD_SO_VAL;
+			n->s[i].uval = PATTERN_LD_SO_VAL;
 		} else if (strcmp(tmp, PATTERN_NORMAL_STR) == 0) {
-			n->s[i].val = PATTERN_NORMAL_VAL;
+			n->s[i].uval = PATTERN_NORMAL_VAL;
 		} else if (strcmp(tmp, PATTERN_STATIC_STR) == 0) {
-			n->s[i].val = PATTERN_STATIC_VAL;
+			n->s[i].uval = PATTERN_STATIC_VAL;
 		} else if (strcmp(tmp, PATTERN_LD_PRELOAD_STR) == 0) {
-			n->s[i].val = PATTERN_LD_PRELOAD_VAL;
+			n->s[i].uval = PATTERN_LD_PRELOAD_VAL;
 		} else {
 			msg(LOG_ERR,
 				"Unknown pattern value %s in line %d",
@@ -434,7 +506,7 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 		}
 
 		errno = 0;
-		long val = strtol(tmp, NULL, 10);
+		unsigned long val = strtoul(tmp, NULL, 10);
 		if (errno) {
 			msg(LOG_ERR,
 				"Error converting val (%s) in line %d",
@@ -446,7 +518,7 @@ static int assign_subject(lnode *n, int type, const char *ptr2, int lineno)
 					"trust can be set to 1 or 0", lineno);
 				goto free_and_error;
 			}
-			n->s[i].val = val;
+			n->s[i].uval = (unsigned int)val;
 		}
 
 		break;
@@ -548,7 +620,8 @@ static int assign_object(lnode *n, int type, const char *ptr2, int lineno)
 		// strings
 		if (set->type != STRING) {
 			msg(LOG_ERR, "rules: line:%d: assign_object: "
-				"cannot assign INT set %s to the STRING attribute",
+				"cannot assign SIGNED set %s to the STRING "
+				"attribute",
 				lineno, defined_set);
 			goto free_and_error;
 		}
@@ -603,7 +676,7 @@ static int assign_object(lnode *n, int type, const char *ptr2, int lineno)
 	case PATH:
 	case DEVICE:
 	case FTYPE:
-	case SHA256HASH:
+	case FILE_HASH:
 	case FMODE: {
 		if (add_attr_set(name, STRING, &index)) {
 			goto free_and_error;
@@ -655,7 +728,7 @@ static int parse_new_format(lnode *n, int lineno)
 		int type;
 		char *ptr2 = strchr(ptr, '=');
 
-                if (ptr2) {
+		if (ptr2) {
 			*ptr2 = 0;
 			ptr2++;
 			if (state == 0) {
@@ -734,10 +807,49 @@ static int parse_set_line(const char * line, int lineno)
 
 
 	char *ptr, *saved, *tmp = sep + 1;
+	char *values = NULL;
+
+	tmp = fapolicyd_strtrim(tmp);
 
 	int type = STRING;
+	bool numeric_found = false;
 
-	if (isdigit(tmp[0])) type = INT;
+	values = strdup(tmp);
+	if (!values)
+		goto free_and_error;
+
+	char *val_ptr, *val_saved;
+	val_ptr = strtok_r(values, ",", &val_saved);
+	while (val_ptr) {
+		char *token = fapolicyd_strtrim(val_ptr);
+		if (!token || *token == '\0') {
+			val_ptr = strtok_r(NULL, ",", &val_saved);
+			continue;
+		}
+
+		errno = 0;
+		char *endptr = NULL;
+		long long sval = strtoll(token, &endptr, 10);
+		if (errno == 0 && endptr && *endptr == '\0') {
+			numeric_found = true;
+			if (sval < 0)
+				type = SIGNED;
+			else if (type != SIGNED)
+				type = UNSIGNED;
+		} else {
+			type = STRING;
+			numeric_found = false;
+			break;
+		}
+
+		val_ptr = strtok_r(NULL, ",", &val_saved);
+	}
+
+	if (!numeric_found)
+		type = STRING;
+
+	free(values);
+	values = NULL;
 
 	if (add_attr_set(name, type, &index)) {
 		goto free_and_error;
@@ -749,10 +861,15 @@ static int parse_set_line(const char * line, int lineno)
 
 	ptr = strtok_r(tmp, ",", &saved);
 	while (ptr) {
+		ptr = fapolicyd_strtrim(ptr);
+		if (!ptr || *ptr == '\0') {
+			ptr = strtok_r(NULL, ",", &saved);
+			continue;
+		}
 		if (type == STRING) {
 			if (append_str_attr_set(set, ptr))
 				goto free_and_error;
-		} else if (type == INT) {
+		} else if (type == SIGNED) {
 			errno = 0;
 			long val = strtol(ptr, NULL, 10);
 			if (errno) {
@@ -760,9 +877,27 @@ static int parse_set_line(const char * line, int lineno)
 					"Error converting val (%s) in line %d",
 					ptr, lineno);
 				goto free_and_error;
-			} else if (append_int_attr_set(set, (int)val))
+			} else if (append_int_attr_set(set,
+					(int64_t)val))
 				goto free_and_error;
 
+		} else if (type == UNSIGNED) {
+			if (*ptr == '-') {
+				msg(LOG_ERR,
+					"Error converting val (%s) in line %d",
+					ptr, lineno);
+				goto free_and_error;
+			}
+			errno = 0;
+			unsigned long val = strtoul(ptr, NULL, 10);
+			if (errno) {
+				msg(LOG_ERR,
+					"Error converting val (%s) in line %d",
+					ptr, lineno);
+				goto free_and_error;
+			} else if (append_int_attr_set(set,
+					(int64_t)val))
+				goto free_and_error;
 		}
 		ptr = strtok_r(NULL, ",", &saved);
 	}
@@ -1023,6 +1158,14 @@ static int check_dirs(unsigned int i, const char *path)
  *
  *    This means only first two can be counted on. Looking for ld.so.cache
  *    is no good because its almost the last option.
+ *
+ * kworker:
+ *    exe=kworker/u130:6 : path=/usr/bin/cat
+ *    exe=kworker/u130:6 : path=/usr/lib64/ld-linux-x86-64.so.2
+ *    exe=/usr/bin/cat : path=/etc/ld.so.cache
+ *    exe=/usr/bin/cat : path=/usr/lib64/libc.so.6
+ *
+ *    Springs to life without ever being an object. Becomes STATE_NORMAL.
  */
 
 // Returns 0 if no match, 1 if a match, -1 on error
@@ -1074,15 +1217,14 @@ static int subj_pattern_test(const subject_attr_t *s, event_t *e)
 			// First thing is ld.so when its used - detected above
 			pinfo->state = STATE_LD_SO;
 		msg(LOG_DEBUG, "pid %d ld.so early path1=%s path2=%s",
-		    pinfo->pid, pinfo->path1 ? pinfo->path1 : "(null)",
-		    pinfo->path2 ? pinfo->path2 : "(null)");
+		    pinfo->pid, pinfo->path1, pinfo->path2);
 		} else    // To get here, pgm matched path1
 			pinfo->state = STATE_NORMAL;
 	}
 
 	// Make a decision
 make_decision:
-	switch (s->val)
+	switch (s->uval)
 	{
 		case PATTERN_NORMAL_VAL:
 			if (pinfo->state == STATE_NORMAL)
@@ -1132,6 +1274,7 @@ static int check_access(const lnode *r, const event_t *e)
 
 
 // Returns 0 if no match, 1 if a match, -1 on error
+__attribute__((hot))
 static int check_subject(lnode *r, event_t *e)
 {
 	unsigned int cnt = 0;
@@ -1159,13 +1302,30 @@ static int check_subject(lnode *r, event_t *e)
 
 		// numbers -> multiple value
 		case AUID:
-		case UID:
-		case SESSIONID:
-		case PID:
-		case PPID: {
-			if (!check_int_attr_set(r->s[cnt].set, subj->val))
+		case SESSIONID: {
+			if (!check_int_attr_set(r->s[cnt].set,
+						(int64_t)subj->uval))
 				return 0;
 			break;
+		}
+		case UID:
+			/*
+			 * A process can present multiple UID values (real,
+			 * effective, saved, filesystem).  Require the rule's
+			 * UID set to intersect the complete credential set the
+			 * subject cached so that any matching identity
+			 * authorizes the rule.
+			 */
+			if (!avl_intersection(&(r->s[cnt].set->tree),
+					      &(subj->set->tree)))
+				return 0;
+			break;
+		case PID:
+		case PPID: {
+			if (!check_int_attr_set(r->s[cnt].set,
+						(int64_t)subj->pid))
+				return 0;
+ 			break;
 		} // case
 
 		// GID is unique in that process can have multiple and
@@ -1193,7 +1353,7 @@ static int check_subject(lnode *r, event_t *e)
 
 		// single value exception
 		case SUBJ_TRUST: {
-			if (subj->val != r->s[cnt].val)
+			if (subj->uval != r->s[cnt].uval)
 				return 0;
 			break;
 		} // case
@@ -1264,6 +1424,7 @@ static int check_subject(lnode *r, event_t *e)
 
 
 // Returns 0 if no match, 1 if a match
+__attribute__((hot))
 static decision_t check_object(lnode *r, event_t *e)
 {
 	unsigned int cnt = 0;
@@ -1308,7 +1469,8 @@ static decision_t check_object(lnode *r, event_t *e)
 		  // skip if fall through
 		  if (type == PATH) {
 			if (r->s[cnt].type == EXE || r->s[cnt].type == EXE_DIR)
-				if (check_str_attr_set(r->s[cnt].set, "untrusted"))
+				if (check_str_attr_set(r->s[cnt].set,
+						       "untrusted"))
 					if (is_obj_trusted(e))
 						return 0;
 		}
@@ -1316,12 +1478,12 @@ static decision_t check_object(lnode *r, event_t *e)
 		// fall through
 
 		case DEVICE:
-		case SHA256HASH:
+		case FILE_HASH:
 		case FMODE: {
 
 			if (!obj->o) {
-				// Treat errors as a denial for sha256
-				if (type == SHA256HASH)
+				// Treat errors as denial for file hash lookups
+				if (type == FILE_HASH)
 					return 0;
 				break;
 			}
@@ -1375,6 +1537,7 @@ static decision_t check_object(lnode *r, event_t *e)
 }
 
 
+__attribute__((hot))
 decision_t rule_evaluate(lnode *r, event_t *e)
 {
 	int d;
@@ -1500,4 +1663,19 @@ void rules_clear(llist *l)
 	l->head = NULL;
 	l->cur = NULL;
 	l->cnt = 0;
+
+	proc_status_mask = 0;
 }
+
+/*
+ * rules_get_proc_status_mask - Report /proc status fields needed by rules.
+ *
+ * Return: bitmap of PROC_STAT_* values observed while parsing the current
+ * rule set. The mask guides process attribute collection so we only read
+ * /proc/<pid>/status once for all requested fields.
+ */
+unsigned int rules_get_proc_status_mask(void)
+{
+       return proc_status_mask;
+}
+
