@@ -94,11 +94,18 @@ int subject_add(s_array *a, const subject_attr_t *subj)
 				return 1;
 			newnode->type = t;
 			if (subj->type >= COMM)
-				newnode->str = subj->str;
-			else if (subj->type == GID)
+			newnode->str = subj->str;
+			else if (subj->type == GID || subj->type == UID)
+			/*
+			 * Attribute sets are reference-count-less blobs.  The
+			 * caller allocates a fresh set for us to adopt so we
+			 * simply transfer ownership to the cached node.
+			 */
 				newnode->set = subj->set;
+			else if (subj->type == PID || subj->type == PPID)
+				newnode->pid = subj->pid;
 			else
-				newnode->val = subj->val;
+				newnode->uval = subj->uval;
 		} else
 			return 1;
 	} else
@@ -142,7 +149,11 @@ void subject_clear(s_array* a)
 		current = a->subj[i];
 		if (current == NULL)
 			continue;
-		if (current->type == GID) {
+		if (current->type == GID || current->type == UID) {
+			/*
+			 * GID/UID attributes borrow dynamically allocated
+			 * sets; release both the set contents and container.
+			 */
 			destroy_attr_set(current->set);
 			free(current->set);
 		} else if (current->type >= COMM)
@@ -168,7 +179,11 @@ void subject_reset(s_array *a, subject_type_t t)
 		subject_attr_t *current = a->subj[t - SUBJ_START];
 		if (current == NULL)
 			return;
-		if (current->type == GID) {
+		if (current->type == GID || current->type == UID) {
+			/*
+			 * GID/UID attributes borrow dynamically allocated
+			 * sets; release both the set contents and container.
+			 */
 			destroy_attr_set(current->set);
 			free(current->set);
 		} else if (current->type >= COMM)
