@@ -1,14 +1,13 @@
 #include <stddef.h>
 #include <string.h>
+#include <stdatomic.h>
 
 #include "backend-manager.h"
 #include "conf.h"
 #include "config.h"
 #include "message.h"
 
-#ifdef USE_RPM
-unsigned int debug_mode;
-#endif
+extern atomic_bool stop;
 
 int main(int argc, char* const argv[]) {
   set_message_mode(MSG_STDERR, DBG_YES);
